@@ -5,7 +5,7 @@
 
     FLINT is free software: you can redistribute it and/or modify it under
     the terms of the GNU Lesser General Public License (LGPL) as published
-    by the Free Software Foundation; either version 2.1 of the License, or
+    by the Free Software Foundation; either version 3 of the License, or
     (at your option) any later version.  See <https://www.gnu.org/licenses/>.
 */
 
@@ -36,7 +36,7 @@
 # include "gc.h"
 #endif
 
-#ifdef FLINT_WANT_ASSERT
+#if FLINT_WANT_ASSERT
 # include <assert.h>
 #endif
 
@@ -57,11 +57,6 @@
 # endif
 #endif
 
-#ifndef __GNUC__
-# define __asm__     asm
-# define __inline__  inline
-#endif
-
 #ifdef va_start
 # define FLINT_HAVE_VA_LIST
 #endif
@@ -73,35 +68,25 @@
 # define FLINT_HAVE_FILE
 #endif
 
-#ifdef FLINT_HAVE_CONSTANT_P
-# define FLINT_CONSTANT_P __builtin_constant_p
-#else
-# define FLINT_CONSTANT_P(x) 0
-#endif
-
 #ifdef FLINT_INLINES_C
 # define FLINT_INLINE
 #else
-# define FLINT_INLINE static __inline__
+# define FLINT_INLINE static inline
 #endif
 
 #ifdef __cplusplus
 extern "C" {
 #endif
 
-/* flint version number */
+/* FLINT version number */
 
 #define __FLINT_VERSION 3
-#define __FLINT_VERSION_MINOR 0
+#define __FLINT_VERSION_MINOR 1
 #define __FLINT_VERSION_PATCHLEVEL 0
-#define FLINT_VERSION ""
+#define FLINT_VERSION "3.1.0"
 #define __FLINT_RELEASE_NUM(a,b,c) ((a)*10000 + (b)*100 + (c))
 #define __FLINT_RELEASE __FLINT_RELEASE_NUM(__FLINT_VERSION, __FLINT_VERSION_MINOR, __FLINT_VERSION_PATCHLEVEL)
 
-#if __GNU_MP_VERSION < 5
-#error GMP 5.0.0 or MPIR 2.6.0 or later are required
-#endif
-
 FLINT_DLL extern char flint_version[];
 
 struct __FLINT_FILE;
@@ -111,20 +96,17 @@ typedef struct __FLINT_FILE FLINT_FILE;
 #define slong mp_limb_signed_t
 #define flint_bitcnt_t ulong
 
-#ifdef FLINT_WANT_ASSERT
-#define FLINT_ASSERT(param) assert(param)
+#if FLINT_WANT_ASSERT
+# define FLINT_ASSERT(param) assert(param)
 #else
-#define FLINT_ASSERT(param)
+# define FLINT_ASSERT(param)
 #endif
 
-#include "longlong.h"
-
 #if defined(__GNUC__)
-# define FLINT_FORCE_INLINE static __attribute__((always_inline)) __inline__
+# define FLINT_FORCE_INLINE static __attribute__((always_inline)) inline
 # define FLINT_STATIC_NOINLINE static __attribute__((noinline))
 # define FLINT_UNUSED(x) UNUSED_ ## x __attribute__((unused))
 # define FLINT_SET_BUT_UNUSED(x) x __attribute__((unused))
-# define FLINT_NORETURN __attribute__ ((noreturn))
 # define FLINT_CONST __attribute__ ((const))
 # define FLINT_WARN_UNUSED __attribute__((warn_unused_result))
 # define FLINT_UNLIKELY(x) __builtin_expect((x),0)
@@ -135,37 +117,50 @@ typedef struct __FLINT_FILE FLINT_FILE;
 # define FLINT_OPTIMIZE_NESTED_2(part) FLINT_OPTIMIZE_NESTED_3(GCC optimize part)
 # define FLINT_OPTIMIZE_NESTED_1(part) FLINT_OPTIMIZE_NESTED_2(#part)
 # define FLINT_OPTIMIZE(x) FLINT_OPTIMIZE_NESTED_1(x)
+# define FLINT_CONSTANT_P __builtin_constant_p
+# define FLINT_UNREACHABLE __builtin_unreachable()
 #else
 # define __attribute__(x)
-# define FLINT_FORCE_INLINE static __inline__
+# if defined(_MSC_VER)
+#  define FLINT_FORCE_INLINE static __forceinline
+#  define FLINT_UNREACHABLE __assume(0)
+# else
+#  define FLINT_FORCE_INLINE static inline
+#  define FLINT_UNREACHABLE
+# endif
 # define FLINT_STATIC_NOINLINE static
 # define FLINT_UNUSED(x) x
 # define FLINT_SET_BUT_UNUSED(x) x
 # define FLINT_WARN_UNUSED
-# define FLINT_NORETURN
 # define FLINT_CONST
 # define FLINT_UNLIKELY(x) (x)
 # define FLINT_LIKELY(x) (x)
 # define FLINT_PUSH_OPTIONS
 # define FLINT_POP_OPTIONS
 # define FLINT_OPTIMIZE(x)
+# define FLINT_CONSTANT_P(x) 0
+#endif
+
+#if defined(__cplusplus)
+# define FLINT_NORETURN [[noreturn]]
+#elif __STDC_VERSION__ < 202300L
+# define FLINT_NORETURN _Noreturn
+#else
+# define FLINT_NORETURN noreturn
 #endif
 
 #if FLINT_USES_TLS
-# if defined(__GNUC__)
-#  define FLINT_TLS_PREFIX __thread
-# elif __STDC_VERSION__ >= 201112L
+# if __STDC_VERSION__ < 202300L && !defined(__cplusplus)
 #  define FLINT_TLS_PREFIX _Thread_local
-# elif defined(_MSC_VER)
-#  define FLINT_TLS_PREFIX __declspec(thread)
 # else
-#  error "thread local prefix defined in C11 or later"
+#  define FLINT_TLS_PREFIX thread_local
 # endif
 #else
 # define FLINT_TLS_PREFIX
 #endif
 
 #if defined(_LONG_LONG_LIMB)
+# define _WORD_FMT "ll"
 # define WORD_FMT "%ll"
 # define WORD_WIDTH_FMT "%*ll"
 # define WORD(xx) (xx##LL)
@@ -177,6 +172,7 @@ typedef struct __FLINT_FILE FLINT_FILE;
 #  define WORD_MIN LLONG_MIN
 # endif
 #else
+# define _WORD_FMT "l"
 # define WORD_FMT "%l"
 # define WORD_WIDTH_FMT "%*l"
 # define WORD(xx) (xx##L)
@@ -198,28 +194,55 @@ typedef struct __FLINT_FILE FLINT_FILE;
 # define FLINT_D_BITS 31
 #endif
 
+#include "longlong.h"
+
+/* memory ********************************************************************/
+
 void * flint_malloc(size_t size);
 void * flint_realloc(void * ptr, size_t size);
 void * flint_calloc(size_t num, size_t size);
 void flint_free(void * ptr);
+void * flint_aligned_alloc(size_t alignment, size_t size);
+void flint_aligned_free(void * ptr);
 
 typedef void (*flint_cleanup_function_t)(void);
 void flint_register_cleanup_function(flint_cleanup_function_t cleanup_function);
 void flint_cleanup(void);
 void flint_cleanup_master(void);
 
-void __flint_set_memory_functions(void *(*alloc_func) (size_t),
-     void *(*calloc_func) (size_t, size_t), void *(*realloc_func) (void *, size_t),
-                                                              void (*free_func) (void *));
+void __flint_set_all_memory_functions(
+        void * (* alloc_func)(size_t),
+        void * (* calloc_func)(size_t, size_t),
+        void * (* realloc_func)(void *, size_t),
+        void (* free_func)(void *),
+        void * (* aligned_alloc_func)(size_t, size_t),
+        void (* aligned_free_func)(void *));
 
-void __flint_get_memory_functions(void *(**alloc_func) (size_t),
-     void *(**calloc_func) (size_t, size_t), void *(**realloc_func) (void *, size_t),
-                                                              void (**free_func) (void *));
+void __flint_get_all_memory_functions(
+        void * (** alloc_func)(size_t),
+        void * (** calloc_func)(size_t, size_t),
+        void * (** realloc_func)(void *, size_t),
+        void (** free_func)(void *),
+        void * (** aligned_alloc_func)(size_t, size_t),
+        void (** aligned_free_func)(void *));
+
+void __flint_set_memory_functions(
+        void * (* alloc_func)(size_t),
+        void * (* calloc_func)(size_t, size_t),
+        void * (* realloc_func)(void *, size_t),
+        void (* free_func)(void *));
+
+void __flint_get_memory_functions(
+        void * (** alloc_func)(size_t),
+        void * (** calloc_func)(size_t, size_t),
+        void * (** realloc_func)(void *, size_t),
+        void (** free_func)(void *));
 
 FLINT_NORETURN void flint_abort(void);
 
-/* If user want to set change abort function from `abort' */
-void flint_set_abort(FLINT_NORETURN void (*func)(void));
+void flint_set_abort(void (*func)(void));
+
+/* threads *******************************************************************/
 
 int flint_get_num_threads(void);
 void flint_set_num_threads(int num_threads);
@@ -229,6 +252,10 @@ void flint_reset_num_workers(int max_workers);
 int flint_set_thread_affinity(int * cpus, slong length);
 int flint_restore_thread_affinity(void);
 
+typedef int thread_pool_handle;
+
+/* randomization and test utilities ******************************************/
+
 FLINT_CONST double flint_test_multiplier(void);
 
 typedef struct
@@ -318,54 +345,14 @@ ulong n_randtest_not_zero(flint_rand_t);
    flint_randclear(xxx); \
    flint_cleanup_master();
 
+/* comparison macros *********************************************************/
+
 #define FLINT_MAX(x, y) ((x) > (y) ? (x) : (y))
 #define FLINT_MIN(x, y) ((x) > (y) ? (y) : (x))
 #define FLINT_ABS(x) ((slong)(x) < 0 ? (-(x)) : (x))
 #define FLINT_SIGN_EXT(x) (-(ulong)((slong)(x) < 0))
 #define FLINT_SGN(x) ((0 < (slong)(x)) - ((slong)(x) < 0))
-
-#define MP_PTR_SWAP(x, y) \
-    do { \
-        mp_limb_t * __txxx; \
-        __txxx = x; \
-        x = y; \
-        y = __txxx; \
-    } while (0)
-
-#define SLONG_SWAP(A, B)    \
-    do {                    \
-        slong __t_m_p_ = A; \
-        A = B;              \
-        B = __t_m_p_;       \
-    } while (0)
-
-#define ULONG_SWAP(A, B)    \
-    do {                    \
-        ulong __t_m_p_ = A; \
-        A = B;              \
-        B = __t_m_p_;       \
-    } while (0)
-
-#define MP_LIMB_SWAP(A, B)      \
-    do {                        \
-        mp_limb_t __t_m_p_ = A; \
-        A = B;                  \
-        B = __t_m_p_;           \
-    } while (0)
-
-#define DOUBLE_SWAP(A, B)    \
-    do {                     \
-        double __t_m_p_ = A; \
-        A = B;               \
-        B = __t_m_p_;        \
-    } while (0)
-
-#define PTR_SWAP(T, A, B)    \
-    do {                    \
-        T* __t_m_p_ = A; \
-        A = B;              \
-        B = __t_m_p_;       \
-    } while (0)
+#define FLINT_SWAP(T, x, y) do { T _swap_t = (x); (x) = (y); (y) = _swap_t; } while(0)
 
 #define r_shift(in, shift) \
     ((shift == FLINT_BITS) ? WORD(0) : ((in) >> (shift)))
@@ -374,7 +361,7 @@ ulong n_randtest_not_zero(flint_rand_t);
     ((shift == FLINT_BITS) ? WORD(0) : ((in) << (shift)))
 
 /* Beware when using the unsigned return value in signed arithmetic */
-static __inline__
+static inline
 mp_limb_t FLINT_BIT_COUNT(mp_limb_t x)
 {
    mp_limb_t zeros = FLINT_BITS;
@@ -386,6 +373,8 @@ mp_limb_t FLINT_BIT_COUNT(mp_limb_t x)
 
 #define FLINT_CLOG2(k)  FLINT_BIT_COUNT((k) - 1)
 
+/* mpn macros ****************************************************************/
+
 #define flint_mpn_zero(xxx, nnn) \
     do \
     { \
@@ -416,11 +405,11 @@ mp_limb_t FLINT_BIT_COUNT(mp_limb_t x)
          (xxx)[ixxx] = yyy; \
    } while (0)
 
-/* common usage of flint_malloc */
+/* allocation macros *********************************************************/
+
 #define FLINT_ARRAY_ALLOC(n, T) (T *) flint_malloc((n)*sizeof(T))
 #define FLINT_ARRAY_REALLOC(p, n, T) (T *) flint_realloc(p, (n)*sizeof(T))
 
-/* temporary allocation */
 #define TMP_INIT \
    typedef struct __tmp_struct { \
       void * block; \
@@ -456,48 +445,59 @@ mp_limb_t FLINT_BIT_COUNT(mp_limb_t x)
       __tmp_root = __tmp_root->next; \
    }
 
-/* compatibility between gmp and mpir */
-#ifndef mpn_com_n
-#define mpn_com_n mpn_com
-#endif
-
-#ifndef mpn_neg_n
-#define mpn_neg_n mpn_neg
-#endif
+/* I/O ***********************************************************************/
 
 int parse_fmt(int * floating, const char * fmt);
 
-int flint_printf(const char * str, ...); /* flint version of printf */
-int flint_sprintf(char * s, const char * str, ...); /* flint version of sprintf */
-int flint_scanf(const char * str, ...); /* flint version of scanf */
-int flint_sscanf(const char * s, const char * str, ...); /* flint version of sscanf */
+int flint_printf(const char * str, ...);
+int flint_sprintf(char * s, const char * str, ...);
+int flint_scanf(const char * str, ...);
+int flint_sscanf(const char * s, const char * str, ...);
 
 #ifdef FLINT_HAVE_VA_LIST
-int flint_vprintf(const char * str, va_list ap); /* va_list version of flint_printf */
+int flint_vprintf(const char * str, va_list ap);
 #endif
 
 #ifdef FLINT_HAVE_FILE
-int flint_fprintf(FILE * f, const char * str, ...); /* flint version of fprintf */
-int flint_fscanf(FILE * f, const char * str, ...); /* flint version of fscanf */
+int flint_fprintf(FILE * f, const char * str, ...);
+int flint_fscanf(FILE * f, const char * str, ...);
 #endif
 
+#if defined(FLINT_HAVE_VA_LIST) && defined(FLINT_HAVE_FILE)
+int flint_vfprintf(FILE * f, const char * str, va_list ap);
+#endif
+
+/* exceptions ****************************************************************/
+
+typedef enum
+{
+   FLINT_ERROR,     /* general error */
+   FLINT_OVERFLOW,  /* overflow */
+   FLINT_IMPINV,    /* impossible inverse */
+   FLINT_DOMERR,    /* domain error */
+   FLINT_DIVZERO,   /* divide by zero */
+   FLINT_EXPOF,     /* exponent overflow */
+   FLINT_INEXACT,   /* inexact error */
+   FLINT_TEST_FAIL  /* test fail */
+} flint_err_t;
+
+FLINT_NORETURN void flint_throw(flint_err_t exc, const char * msg, ...);
+
+/* checked multiplication ****************************************************/
+
 FLINT_INLINE slong flint_mul_sizes(slong x, slong y)
 {
     ulong hi, lo;
 
     umul_ppmm(hi, lo, (ulong) x, (ulong) y);
+
     if (hi != 0 || lo > WORD_MAX)
-    {
-        flint_printf("Exception (flint). Overflow creating size %wd x %wd object.\n", x, y);
-        flint_abort();
-    }
+        flint_throw(FLINT_OVERFLOW, "Overflow creating size %wd x %wd object.\n", x, y);
+
     return lo;
 }
 
-#include "exception.h"
-
-/* Defined ahead of nmod.h, fmpz.h and fmpq.h so that the types are usable in
- * prototypes without importing those headers */
+/* FLINT generic type definitions ********************************************/
 
 typedef struct
 {
@@ -522,6 +522,21 @@ typedef fmpq fmpq_t[1];
 #define fmpq_numref(__x) (&(__x)->num)
 #define fmpq_denref(__y) (&(__y)->den)
 
+/* fmpz macros ***************************************************************/
+
+/* The largest bit count for an fmpz to be small */
+#define SMALL_FMPZ_BITCOUNT_MAX (FLINT_BITS - 2)
+
+/* Minimum and maximum value for a small fmpz */
+#define COEFF_MIN (-((WORD(1) << SMALL_FMPZ_BITCOUNT_MAX) - WORD(1)))
+#define COEFF_MAX ((WORD(1) << SMALL_FMPZ_BITCOUNT_MAX) - WORD(1))
+
+/* Conversions between mpz_ptr and fmpz_t */
+#define PTR_TO_COEFF(x) (((ulong) (x) >> 2) | (WORD(1) << (FLINT_BITS - 2)))
+#define COEFF_TO_PTR(x) ((mpz_ptr) (((ulong)x) << 2))
+
+#define COEFF_IS_MPZ(x) (((x) >> SMALL_FMPZ_BITCOUNT_MAX) == WORD(1))
+
 #ifdef __cplusplus
 }
 #endif
